import numpy as np
import dash_ag_grid as dag
from dash import dcc, html
import dash_bootstrap_components as dbc

# Sliders
marks = {0: '0.0', 1: '1.0'}
marks.update({i: f"{i:.1f}" for i in np.linspace(0.1, 0.9, 9)})

for k, v in marks.copy().items():
    marks[k] = {'label': v, 'style': {'fontSize': '1em'}}

sliders = dbc.Stack(
    [
        dbc.Label('Negative Electrode: [0.00, 1.00]',
                  id='neg-slider-label', class_name='bold-label mt-0'),
        dcc.RangeSlider(
            id='neg-slider', updatemode='drag',
            min=0.0, max=1.0, step=0.01, value=[0., 1], marks=marks,
        ),

        dbc.Label('Positive Electrode: [0.00, 1.00]',
                  id='pos-slider-label', class_name='bold-label mt-3'),
        dcc.RangeSlider(
            id='pos-slider', updatemode='drag',
            min=0.0, max=1.0, step=0.01, value=[0, 1], marks=marks,
        ),
    ],
    class_name='mx-auto',
    style={'maxWidth': '900px'},
)

# Optimization buttons
optimize_btns = dbc.Stack(
    [
        dbc.Button(
            'Coarse Search',
            id='coarse-btn',
            class_name='amp-btn',
        ),
        dbc.Button(
            'Minimize Error',
            id='min-err-btn',
            class_name='amp-btn',
        ),
    ],
    gap=3,
    direction='horizontal',
    style={'maxWidth': '900px'},
    class_name='justify-content-end my-3 mx-auto',
)

page_spinner = dbc.Spinner(
    size='lg',
    color='white',
    fullscreen=True,
    id='page-spinner',
    children=html.Div(id='spinner-div'),
    fullscreen_style={
        'zIndex': '2000',
        'backgroundColor': 'rgba(0, 0, 0, 0.5)',
    },
)

# Terminal output
terminal = dbc.Row(
    [
        dcc.Markdown(
            id='terminal-out',
            className='dbc markdown',
        ),
    ],
    style={'maxWidth': '900px'},
    class_name='pb-3 pt-0 px-0 m-0 mx-auto',
)

# Log toast
log_toast = dbc.Toast(
    is_open=False,
    duration=3000,
    icon='success',
    id='log-toast',
    header='Successful new log',
    style={'position': 'fixed', 'top': 66, 'right': 10, 'width': 350},
)

# Logging buttons
download = dcc.Download(id="download-csv")
logging_btns = dbc.Stack(
    [
        dbc.Button('Log Output', id='add-row-btn',
                   class_name='log-btn success'),
        dbc.Button('Delete Selected', id='delete-btn',
                   class_name='log-btn danger'),
        dbc.Button('Export Data', id='export-btn',
                   class_name='log-btn secondary'),
    ],
    gap=3,
    direction='horizontal',
    style={'maxWidth': '900px'},
    class_name='justify-content-end mx-auto mb-3',
)

# Logging table
column_defs = [
    {
        'minWidth': 125,
        'field': 'filename',
        'headerName': 'filename',
        'checkboxSelection': True,
        'headerCheckboxSelection': True,
    },
    *[
        {'headerName': f"{c}", 'field': f"{c}"}
        for c in ['x0_neg', 'x100_neg', 'x0_pos', 'x100_pos', 'iR', 'fun']
    ],
]

logging_table = dbc.Row(
    dag.AgGrid(
        rowData=[],
        id='ag-grid',
        columnDefs=column_defs,
        defaultColDef={
            'flex': 1,
            'minWidth': 80,
            'sortable': False,
        },
        dashGridOptions={
            'domLayout': 'normal',
            'rowSelection': 'multiple',
        },
    ),
    style={'maxWidth': '900px'},
    class_name='amp-ag-grid mx-auto mb-3',
)



import os
import webbrowser

import dash
import dash_bootstrap_components as dbc
from dash_breakpoints import WindowBreakpoints
from dash import dcc, html, Output, Input, State
from dash_extensions.pages import setup_page_components

from ampworks.dqdv.gui_files.layout.navbar import navbar
from ampworks.dqdv.gui_files.layout.sidebar import sidebar

here = os.path.dirname(__file__)
pages_folder = os.path.join(here, 'pages')

dbc_css = 'https://cdn.jsdelivr.net/gh/AnnMarieW/' \
    + 'dash-bootstrap-templates/dbc.min.css'

app = dash.Dash(
    __name__,
    use_pages=True,
    title='ampworks',
    pages_folder=pages_folder,
    external_stylesheets=[
        dbc_css,
        dbc.themes.BOOTSTRAP,
        dbc.icons.FONT_AWESOME,
    ],
)

breakpoints = WindowBreakpoints(
    id='breakpoints',
    widthBreakpointNames=['sm', 'lg'],
    widthBreakpointThresholdsPx=[1199],
)

dummy_block = html.Div(id='dummy-block', className='dummy-block')

main_page = html.Div(
    id='main-page',
    className='main-page mx-auto my-3',
    children=[dash.page_container, setup_page_components()],
)

dash_link = html.A(
    className='link',
    children='dash 2.18',
    href='https://dash.plotly.com/',
)

footer = html.Div(
    [
        html.Div('© Alliance for Sustainable Energy, LLC'),
        html.Div(['Created using ', dash_link], className='text-end'),
    ],
    id='footer',
    className='footer px-4',
)

grid_container = html.Div(
    id='grid-container',
    className='grid-container',
    children=[dummy_block, main_page, footer],
)

app.layout = html.Div(
    [
        dcc.Location(id='url'),
        breakpoints,
        navbar,
        sidebar,
        grid_container,
    ],
)

# Transform page on width breakpoint
app.clientside_callback(
    """
    function trasformPage(flag, width) {

        if (width < 1200) {
            return [false, false, {}, true, true, "Control Panel"];
        }

        sidebarStyle = {
            "top": "4em",
            "position": "fixed",
        };

        return [false, true, sidebarStyle, false, false, ""];
    }
    """,
    Output('navlinks-collapse', 'is_open', allow_duplicate=True),
    Output('sidebar', 'is_open', allow_duplicate=True),
    Output('sidebar', 'style'),
    Output('sidebar', 'backdrop'),
    Output('sidebar', 'close_button'),
    Output('sidebar', 'title'),
    Input('breakpoints', 'widthBreakpoint'),
    State('breakpoints', 'width'),
    prevent_initial_call=True,
)


def run(debug: bool = False) -> None:

    if os.environ.get("WERKZEUG_RUN_MAIN") != "true":
        webbrowser.open_new("http://127.0.0.1:8050/")

    app.run(
        debug=debug,
        dev_tools_silence_routes_logging=True,
    )



"""
TODO
----

"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import scipy.optimize as opt
import scipy.interpolate as interp
from scipy.signal import savgol_filter
import matplotlib.pyplot as plt

if TYPE_CHECKING:  # pragma: no cover
    import numpy.typing as npt


class Fitter:

    def __init__(self, df_neg: pd.DataFrame = None, df_pos: pd.DataFrame = None,
                 df_cell: pd.DataFrame = None, **kwargs) -> None:
        """
        Wrapper class for differential analysis.

        Parameters
        ----------
        df_neg : pd.DataFrame
            Data for negative electrode OCV.
        df_pos : pd.DataFrame
            Data for positive electrode OCV.
        df_cell : pd.DataFrame
            Data for full cell OCV and derivatives.
        **kwargs : dict, optional
            Keyword arguments. Optional key/value pairs are below:

            ============ ==================================================
            Key          Value (*type*, default)
            ============ ==================================================
            smoothing    smoothing window for fits (*int*, 3)
            figure_font  fontsize for figure elements (*int*, 10)
            bounds       +/- bounds (*list[float]*, [0.1] * 4)
            maxiter      maximum fit iterations (*int*, 1e5)
            xtol         optimization tolerance on x (*float*, 1e-9)
            cost_terms   terms in err func (*list[str]*, ['dqdv', 'dvdq'])
            ============ ==================================================

        Raises
        ------
        ValueError
            Invalid keyword arguments.

        Notes
        -----
        * The df_neg, df_pos, and df_cell dataframes are all required inputs.
          The default ``None`` values allow you to initialize the class first
          and then add each dataframe one at a time. This is primarily so the
          class interfaces well with the data loader in the GUI.
        * Bound indices correspond to x0_neg, x100_neg, x0_pos, and x100_pos.
          Set bounds[i] equal to 1 to use the full interval [0, 1] for x[i].
        * If x[i] +/- bounds[i] exceeds the limits [0, 1], the lower and/or
          upper bounds will be corrected to 0 and/or 1, respectively.
        * The cost_terms list must be a subset of {'voltage', 'dqdv', 'dvdq'}.
          When 'voltage' is included, an iR term is fit in addition to the
          x0/x100 terms. Otherwise, iR is forced towards zero.

        """

        for attr in ['_df_neg', '_df_pos', '_df_cell']:
            setattr(self, attr, None)

        if df_neg is not None:
            self.df_neg = df_neg

        if df_pos is not None:
            self.df_pos = df_pos

        if df_cell is not None:
            self.df_cell = df_cell

        self.smoothing = kwargs.pop('smoothing', 3)
        self.figure_font = kwargs.pop('figure_font', 10)
        self.bounds = kwargs.pop('bounds', [0.1] * 4)
        self.maxiter = kwargs.pop('maxiter', 1e5)
        self.xtol = kwargs.pop('xtol', 1e-9)
        self.cost_terms = kwargs.pop('cost_terms', ['dqdv', 'dvdq'])

        if len(kwargs) != 0:
            invalid_keys = list(kwargs.keys())
            raise ValueError(f"Invalid keyword arguments: {invalid_keys}")

    @property
    def df_neg(self) -> pd.DataFrame:
        """
        Get or set the negative electrode dataframe.

        Columns must include both 'soc' for state of charge and 'voltage' for
        the half-cell voltage. 'soc' should be normalized to [0, 1].

        """
        return self._df_neg

    @df_neg.setter
    def df_neg(self, value: pd.DataFrame) -> None:
        if not isinstance(value, pd.DataFrame):
            raise TypeError("df_neg must be type pd.DataFrame.")
        elif 'soc' not in value.columns:
            raise ValueError("'soc' is missing from df_neg.columns.")
        elif 'voltage' not in value.columns:
            raise ValueError("'voltage' is missing from df_neg.columns.")

        self._df_neg = value
        self.ocv_neg = self.ocv_spline(self._df_neg, 'neg')

    @property
    def df_pos(self) -> pd.DataFrame:
        """
        Get or set the positive electrode dataframe.

        Columns must include both 'soc' for state of charge and 'voltage' for
        the half-cell voltage. 'soc' should be normalized to [0, 1].

        """
        return self._df_pos

    @df_pos.setter
    def df_pos(self, value: pd.DataFrame) -> None:
        if not isinstance(value, pd.DataFrame):
            raise TypeError("df_pos must be type pd.DataFrame.")
        elif 'soc' not in value.columns:
            raise ValueError("'soc' is missing from df_pos.columns.")
        elif 'voltage' not in value.columns:
            raise ValueError("'voltage' is missing from df_pos.columns.")

        self._df_pos = value
        self.ocv_pos = self.ocv_spline(self._df_pos, 'pos')

    @property
    def df_cell(self) -> pd.DataFrame:
        """
        Get or set the full cell dataframe.

        Columns must include 'soc' for state of charge, 'voltage' for the cell
        voltage, 'dsoc_dV' for the derivative dsov/dV, and 'dV_dsoc' for the
        derivative dV/dsoc. 'soc' should be normalized to [0, 1].

        """
        return self._df_cell

    @df_cell.setter
    def df_cell(self, value: pd.DataFrame) -> None:
        import numpy as np

        if not isinstance(value, pd.DataFrame):
            raise TypeError("df_cell must be type pd.DataFrame.")
        elif 'soc' not in value.columns:
            raise ValueError("'soc' is missing from df_cell.columns.")
        elif 'voltage' not in value.columns:
            raise ValueError("'voltage' is missing from df_cell.columns.")
        elif 'dsoc_dV' not in value.columns:
            raise ValueError("'dsoc_dV' is missing from df_cell.columns.")
        elif 'dV_dsoc' not in value.columns:
            raise ValueError("'dV_dsoc' is missing from df_cell.columns.")

        self._df_cell = value
        output = self.ocv_spline(self._df_cell, 'cell')
        self.ocv_cell, self.dqdv_cell, self.dvdq_cell = output

        self._soc = np.linspace(0., 1., 500)
        self._soc_mid = 0.5 * (self._soc[:-1] + self._soc[1:])

        self.V_dat = self.ocv_cell(self._soc)
        self.dqdv_dat = self.dqdv_cell(self._soc_mid)
        self.dvdq_dat = self.dvdq_cell(self._soc_mid)

    @property
    def smoothing(self) -> int:
        """
        Get or set the fit smoothing.

        The fitted dsoc/dV and dV/dsoc curves often have a lot of noise in
        them because they carry noise over from both half-cell OCV curves.
        This property is used to smooth the fitted derivatives. The smoothed
        curves are used to determine error between the fit and data.

        """
        return self._smoothing

    @smoothing.setter
    def smoothing(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("smoothing must be type int.")

        self._smoothing = value

    @property
    def figure_font(self) -> int:
        """
        Get or set the figure fontsize.

        """
        return self._figure_font

    @figure_font.setter
    def figure_font(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("figure_font must be type int.")

        self._figure_font = value

    @property
    def bounds(self) -> list[float]:
        """
        Get or set the bounds for the constrained fit routine.

        """
        return self._bounds

    @bounds.setter
    def bounds(self, value: list[float]) -> None:
        if not isinstance(value, (list, tuple)):
            raise TypeError("bounds must be an iterable.")

        if len(value) != 4:
            raise ValueError("bounds must have length 4.")

        for v in value:
            if not isinstance(v, (float, int)):
                raise TypeError("All bounds[i] must be type float.")
            elif v < 0.001:
                raise ValueError("All bounds[i] must be >= 0.001.")
            elif v > 1.:
                raise ValueError("All bounds[i] must be <= 1.")

        self._bounds = value

    @property
    def maxiter(self) -> int:
        """
        Get or set the maximum iterations for the constrained fit routine.

        """
        return self._maxiter

    @maxiter.setter
    def maxiter(self, value: int) -> None:
        if isinstance(value, float):
            if not value.is_integer():
                raise TypeError("maxiter must be type int.")
            else:
                value = int(value)

        if not isinstance(value, int):
            raise TypeError("maxiter must be type int.")

        if value <= 0:
            raise ValueError("maxiter must be positive.")

        self._maxiter = value

    @property
    def xtol(self) -> float:
        """
        Get or set the 'x' tolerance for the constrained fit routine.

        """
        return self._xtol

    @xtol.setter
    def xtol(self, value: float) -> None:
        if not isinstance(value, (float, int)):
            raise TypeError("xtol must be type float.")

        if value <= 0.:
            raise ValueError("xtol must be positive.")

        self._xtol = value

    @property
    def cost_terms(self) -> list[str]:
        """
        Get or set which terms are included in the constrained fit's cost
        function. Options are 'voltage', 'dqdv', and/or 'dvdq'.

        """
        return self._cost_terms

    @cost_terms.setter
    def cost_terms(self, value: list[str]) -> None:
        if not isinstance(value, (list, tuple)):
            raise TypeError("cost_terms must be an iterable.")

        for v in value:
            if not isinstance(v, str):
                raise TypeError("All cost_terms[i] must be type str.")
            elif v not in ['voltage', 'dqdv', 'dvdq']:
                raise ValueError(f"Invalid cost term '{v}'. Valid options are"
                                 + " 'voltage', 'dqdv', and/or 'dvdq'.")

        if len(value) == 0:
            raise ValueError("len(cost_terms) == 0. Must include at least one"
                             + " of {'voltage', 'dqdv', 'dvdq'}.")

        self._cost_terms = value

    def ocv_spline(self, df: pd.DataFrame, name: str) -> callable:
        """
        Generate OCV interpolation functions.

        Parameters
        ----------
        df : pd.DataFrame
            Data with 'soc' and 'voltage' columns. If name is 'cell', the
            data should also include 'dsoc_dV' and 'dV_dsoc' columnes.
        name : str
            Dataset name, from {'neg', 'pos', 'cell'}.

        Returns
        -------
        ocv : callable
            If name is not 'cell' return a makima interpolation for OCV.
        ocv, dqdv, dvdq : tuple[callable]
            If name is 'cell', return a makima interpolation for OCV, dqdv,
            and dvdq.

        """

        from ..math import makima

        df = clean_df(df, unique_cols=['soc', 'voltage'], sort_by='soc')

        ocv = interp.CubicSpline(df.soc.values, df.voltage.values)

        if name == 'cell':
            dqdv_cell = makima(df.soc.values, df.dsoc_dV.values)
            dvdq_cell = makima(df.soc.values, df.dV_dsoc.values)
            return ocv, dqdv_cell, dvdq_cell
        else:
            return ocv

    def err_terms(self, params: npt.ArrayLike,
                  full_output: bool = False) -> dict:
        """
        Calculate error between the fit and data.

        Parameters
        ----------
        params : ArrayLike, shape(n,)
            Array for x0_neg, x100_neg, x0_pos, x100_pos, and iR (optional).
        full_output : bool, optional
            Flag to return all data. The default is False.

        Returns
        -------
        errs : tuple[float]
            If full_output is False, return voltage error, dqdv error, and
            dvdq error values.
        full_output : dict
            If full_output is True, return a dictionary with the fit arrays,
            data arrays, and error values.

        Notes
        -----
        Error terms are the mean absolute errors between data and predicted
        values, normalized by the data. The normalization reduces preferences
        to fit any one cost term over others when more than one is considered.
        In addition, normalizing allows the errors to be added when the cost
        function includes more than one term.

        """

        params = np.asarray(params)

        if params.size == 5:
            x0_neg, x100_neg, x0_pos, x100_pos, iR = params
        else:
            x0_neg, x100_neg, x0_pos, x100_pos, iR = *params, 0.

        x_neg = x0_neg + (x100_neg - x0_neg) * self._soc
        x_pos = x0_pos + (x100_pos - x0_pos) * self._soc

        V_dat = self.V_dat
        V_fit = self.ocv_pos(x_pos) - self.ocv_neg(x_neg) - iR

        dqdv_dat = self.dqdv_dat
        dqdv_fit = np.diff(self._soc) / np.diff(V_fit)
        dqdv_fit = savgol_filter(dqdv_fit, self.smoothing, 2, mode='nearest')

        dvdq_dat = self.dvdq_dat
        dvdq_fit = np.diff(V_fit) / np.diff(self._soc)
        dvdq_fit = savgol_filter(dvdq_fit, self.smoothing, 2, mode='nearest')

        err1 = np.mean(np.abs((V_fit - V_dat) / V_dat))
        err2 = np.mean(np.abs((dqdv_fit - dqdv_dat) / dqdv_dat))
        err3 = np.mean(np.abs((dvdq_fit - dvdq_dat) / dvdq_dat))

        output = {
            'err1': err1,
            'err2': err2,
            'err3': err3,
        }

        if full_output:
            output.update({
                'soc': self._soc,
                'soc_mid': self._soc_mid,
                'V_dat': V_dat,
                'V_fit': V_fit,
                'dqdv_dat': dqdv_dat,
                'dqdv_fit': dqdv_fit,
                'dvdq_dat': dvdq_dat,
                'dvdq_fit': dvdq_fit,
            })

        return output

    def err_func(self, params: npt.ArrayLike) -> float:
        """
        The cost function for coarse_search and constrained_fit.

        Parameters
        ----------
        params : ArrayLike, shape(n,)
            Array for x0_neg, x100_neg, x0_pos, x100_pos, and optionally iR.

        Returns
        -------
        err_tot : float
            Total error based on a combination of cost_terms.

        """

        output = self.err_terms(params)

        err = 0.
        if 'voltage' in self.cost_terms:
            err += output['err1']

        if 'dqdv' in self.cost_terms:
            err += output['err2']

        if 'dvdq' in self.cost_terms:
            err += output['err3']

        return err

    def coarse_search(self, Nx: int) -> dict:
        """
        Determine the minimum error by evaluating parameter sets taken from
        intersections of a coarse grid. Parameter sets where x0 < x100 for
        either electrode are ignored.

        Parameters
        ----------
        Nx : int
            Number of discretizations between [0, 1] for each parameter.

        Returns
        -------
        summary : dict
            Summarized results from the coarse search.

        """

        from ..math import combinations

        span = np.linspace(0., 1., Nx)
        names = ['x0_neg', 'x100_neg', 'x0_pos', 'x100_pos']

        params = combinations([span] * 4, names=names)

        valid_ps = []
        for p in params:
            if p['x0_neg'] < p['x100_neg'] and p['x0_pos'] < p['x100_pos']:
                valid_ps.append(p)

        errs = []
        for p in valid_ps:
            values = np.array(list(p.values()))
            errs.append(self.err_func(values))

        index = np.argmin(errs)

        summary = {
            'nfev': len(errs),
            'fun': errs[index],
            'x': np.array(list(valid_ps[index].values())),
            'x_map': list(valid_ps[index].keys()),
        }

        return summary

    def constrained_fit(self, x0: npt.ArrayLike) -> dict:
        """
        Run a trust-constrained local optimization routine to minimize error
        between the fit and data.

        Parameters
        ----------
        x0 : ArrayLike, shape(n,)
            Initial x0_neg, x100_neg, x0_pos, x100_pos, and optionally iR.

        Returns
        -------
        summary : dict
            Summarized results from the optimization routine.

        """

        x0 = np.asarray(x0)

        output = self.err_terms(x0, full_output=True)

        iR0 = (output['V_fit'] - output['V_dat']).mean()

        if x0.size == 5:
            x0[-1] = iR0
        elif x0.size == 4:
            x0 = np.hstack([x0, iR0])

        lower = np.zeros_like(x0)
        upper = np.ones_like(x0)
        for i in range(4):
            lower[i] = max(0., x0[i] - self.bounds[i])
            upper[i] = min(1., x0[i] + self.bounds[i])

        if 'voltage' in self._cost_terms:
            lower[-1] = -np.inf
            upper[-1] = np.inf
        else:
            lower[-1] = 0.
            upper[-1] = 0.

        bounds = [(L, U) for L, U in zip(lower, upper)]

        constr_neg = opt.LinearConstraint([[1, -1, 0, 0, 0]], -np.inf, 0.)
        constr_pos = opt.LinearConstraint([[0, 0, 1, -1, 0]], -np.inf, 0.)
        constr_iR = opt.LinearConstraint([[0, 0, 0, 0, 1]], *bounds[-1])

        constraints = [constr_neg, constr_pos, constr_iR]

        options = {
            'maxiter': self.maxiter,
            'xtol': self.xtol,
        }

        warnings.filterwarnings('ignore')

        result = opt.minimize(self.err_func, x0, method='trust-constr',
                              bounds=bounds, constraints=constraints,
                              options=options)

        warnings.filterwarnings('default')

        keys = ['success', 'message', 'nfev', 'niter', 'fun', 'x']
        summary = dict((k, result[k]) for k in keys)

        summary['x_map'] = ['x0_neg', 'x100_neg', 'x0_pos', 'x100_pos', 'iR']

        return summary

    def plot(self, params: npt.ArrayLike, **kwargs) -> None:
        """
        Plot the fit vs. data.

        Parameters
        ----------
        params : ArrayLike, shape(n,)
            Parameters x0_neg, x100_neg, x0_pos, x100_pos, and optionally iR.
        **kwargs : dict, optional
            Keyword arguments. Optional key/value pairs are below:

            ============== ============================================
            Key            Value (*type*, default)
            ============== ============================================
            fig            1x3 subplot figure to fill (*object*, None)
            voltage_ylims  ylimits for voltage (*list[float]*, None)
            dqdv_ylims     ylimits for dsoc_dV (*list[float]*, None)
            dvdq_ylims     ylimits for dV_dsoc (*list[float]*, None)
            ============== ============================================

        Returns
        -------
        None.

        """

        from ..plotutils import format_ticks, add_text

        fig = kwargs.pop('fig', None)
        figsize = kwargs.pop('figsize', [12., 3.])

        plt.rcParams['font.size'] = self.figure_font

        if fig is None:
            fig, ax = plt.subplots(nrows=1, ncols=3, figsize=figsize,
                                   layout='tight')
        else:
            ax = fig.get_axes()
            _ = [ax[i].clear() for i in range(3)]

        output = self.err_terms(params, full_output=True)

        dat_options = {'color': 'C1', 'linestyle': '', 'alpha': 0.5,
                       'marker': 'o', 'label': 'Data'}

        fit_options = {'color': 'k', 'linestyle': '-', 'label': 'Model'}

        ax[0].set_ylabel('Voltage [V]')
        ax[0].plot(output['soc'], output['V_dat'], **dat_options)
        ax[0].plot(output['soc'], output['V_fit'], **fit_options)
        add_text(ax[0], 0.1, 0.8, f"{output['err1']:.3e}")

        ax[1].set_ylabel(r'$dq/dV$ [V$^{-1}$]')
        ax[1].plot(output['soc_mid'], output['dqdv_dat'], **dat_options)
        ax[1].plot(output['soc_mid'], output['dqdv_fit'], **fit_options)
        add_text(ax[1], 0.6, 0.8, f"{output['err2']:.3e}")

        ax[2].set_ylabel(r'$dV/dq$ [V]')
        ax[2].plot(output['soc_mid'], output['dvdq_dat'], **dat_options)
        ax[2].plot(output['soc_mid'], output['dvdq_fit'], **fit_options)
        add_text(ax[2], 0.6, 0.8, f"{output['err3']:.3e}")

        ax[0].legend(loc='lower right', frameon=False)

        if 'voltage_ylims' in kwargs.keys():
            ax[0].set_ylim(kwargs.get('voltage_ylims'))

        if 'dqdv_ylims' in kwargs.keys():
            ax[1].set_ylim(kwargs.get('dqdv_ylims'))

        if 'dvdq_ylims' in kwargs.keys():
            ax[2].set_ylim(kwargs.get('dvdq_ylims'))

        for i in range(3):
            format_ticks(ax[i])
            ax[i].set_xlabel(r'SOC, $q$ [$-$]')


def clean_df(df: pd.DataFrame, unique_cols: list[str] = [],
             sort_by: str = None) -> pd.DataFrame:
    """
    Clean up dataframes for dqdv analysis.

    Drop all nan values, ensure specified columns do not have duplicates,
    and sort the whole dataframe according to a given column.

    Parameters
    ----------
    df : pd.DataFrame
        A pandas dataframe.
    unique_cols : list[str], optional
        Columns to remove duplicate values, if present. The default is [].
    sort_by : str, optional
        Column name used to sort dataframe. The default is None.

    Returns
    -------
    df : pd.DataFrame
        Cleaned and sorted dataframe.

    """

    for col in unique_cols:
        _, mask = np.unique(df[col], return_index=True)
        df = df.iloc[mask]

    df = df.dropna()
    if sort_by is not None:
        df = df.sort_values(sort_by)

    df = df.reset_index(drop=True)

    return df


def preprocess_df(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    from ..math import dydx, makima

    sw = kwargs.get('smoothing_win', None)
    V_lower = kwargs.get('V_lower', min(df.voltage))
    V_upper = kwargs.get('V_upper', max(df.voltage))
    V_res = kwargs.get('V_res', 1e-3)
    capacity = kwargs.get('capacity', max(abs(df.amphr)))
    flip_amphr = kwargs.get('flip_amphr', False)

    df = clean_df(df, unique_cols=['voltage', 'amphr'], sort_by='voltage')

    data = {}

    voltage = np.arange(V_lower, V_upper, V_res)
    data['voltage'] = voltage

    amphr_func = makima(df.voltage, df.amphr)
    amphr = np.abs(amphr_func(voltage))

    if flip_amphr:
        amphr = capacity - amphr

    data['amphr'] = amphr
    data['soc'] = amphr / capacity

    options = {'smoothing_win_y': sw, 'lead_value': np.nan}
    data['damphr_dV'] = dydx(data['voltage'], data['amphr'], **options)
    data['dsoc_dV'] = dydx(data['voltage'], data['soc'], **options)

    options = {'smoothing_win_x': sw, 'lead_value': np.nan}
    data['dV_damphr'] = dydx(data['amphr'], data['voltage'], **options)
    data['dV_dsoc'] = dydx(data['soc'], data['voltage'], **options)

    if 'time' in df.columns:
        time_func = makima(df.voltage, df.time)
        data['time'] = time_func(data['voltage'])

    if 'current' in df.columns:
        current_func = makima(df.voltage, df.current)
        data['current'] = current_func(data['voltage'])

    output = pd.DataFrame(data)

    return output


def post_process(capacity: npt.ArrayLike, x: npt.ArrayLike) -> dict:
    """
    Determine degradation parameters.

    Uses full cell capacity and fitted x0/x100 values from dqdv/dvdq fits to
    calculate theoretical electrode capacities, loss of active material (LAM),
    and total inventory losses (TIL). TIL is used instead of LLI (loss of
    lithium inventory) because this analysis is also valid for intercalation
    electrodes with active species other than lithium.

    Electrode capacities (Q) and losses of active material (LAM) are

    .. math::

        Q_{ed} = \\frac{\\rm capacity}{x_{100,ed} - x_{0,ed}}, \\quad \\quad
        {\\rm LAM}_{ed} = 1 - \\frac{Q_{ed}}{Q_{ed}[0]},

    where :math:`ed` is used generically 'electrode'. In the output, subscripts
    'neg' and 'pos' are used to differentiate between the negative and positive
    electrodes, respectively. Loss of inventory is

    .. math::

        I = x_{100,neg}Q_{neg} + x_{100,pos}Q_{pos}, \\quad \\quad
        {\\rm TIL} = 1 - \\frac{I}{I[0]},

    where :math:`I` is an array of inventories calculated from the capacities
    :math:`Q` above. The 'offset' output can also sometimes serve as a helpful
    metric. It is simply the difference between 'x0_neg' and 'x0_pos'.

    Parameters
    ----------
    capacity : ArrayLike, shape(n,)
        Full cell capacity values per fitted profile.
    x : ArrayLike, shape(n,4)
        Fitted x0/x100 values. Row i corresponds to capacity[i], with column
        order: x0_neg, x100_neg, x0_pos, x100_pos.

    Raises
    ------
    ValueError
        capacity.size != x.shape[0].
    ValueError
        x.shape[1] != 4.

    Returns
    -------
    results : dict
        Electrode capacities (Q) and loss of active material (LAM) for the
        negative (neg) and positive (pos) electrodes, and total loss of
        intentory (TLI). Capacity units match the ``capacity`` input. All
        other outputs are unitless.

    """

    x = np.asarray(x)
    capacity = np.asarray(capacity)

    if capacity.size != x.shape[0]:
        raise ValueError(f"{capacity.size=} != {x.shape[0]=}.")

    if x.shape[1] == 4:
        x0_neg, x100_neg, x0_pos, x100_pos = x.T
    else:
        raise ValueError(f"{x.shape[1]=} != 4.")

    Q_neg = capacity / (x100_neg - x0_neg)
    Q_pos = capacity / (x100_pos - x0_pos)

    offset = x0_neg - x0_pos

    LAM_neg = 1. - Q_neg / Q_neg[0]
    LAM_pos = 1. - Q_pos / Q_pos[0]

    inventory = x100_neg*Q_neg + (1. - x100_pos)*Q_pos
    TIL = 1. - inventory / inventory[0]

    results = {
        'Q_neg': Q_neg,
        'Q_pos': Q_pos,
        'offset': offset,
        'LAM_neg': LAM_neg,
        'LAM_pos': LAM_pos,
        'TIL': TIL,
    }

    return results


def run_gui() -> None:
    """
    Run a graphical interface for the Fitter class.

    Returns
    -------
    None.

    """

    from .gui_files import _gui

    _gui.run()



import io
import base64

import dash
import numpy as np
import pandas as pd
import ampworks as amp
import plotly.io as pio
import plotly.graph_objects as go
import dash_bootstrap_components as dbc
from dash import dcc, html, Output, Input, State
from dash_bootstrap_templates import load_figure_template

from ampworks.dqdv.gui_files.pages.figures import figure, placeholder_fig

fitter = amp.dqdv.Fitter()
optimal_params = False

load_figure_template(['bootstrap', 'bootstrap_dark'])


def file_upload(label, identifier):
    if identifier == 'neg':
        upload_label = dbc.Label(label, class_name='bold-label')
    else:
        upload_label = dbc.Label(label, class_name='bold-label mt-3')

    horizontal_divider = html.Hr(className='m-0')

    file_uploader = dcc.Upload(
        id={'type': 'upload', 'index': identifier},
        children=html.Div(
            [
                'Drag & Drop or ',
                html.A('Select File'),
            ],
            style={'textAlign': 'center'},
        ),
        multiple=False,
        className='upload-btn',
        className_active='upload-btn-active',
        className_reject='upload-btn-reject',
        accept='.csv, text/csv, application/csv',
        style={
            'width': '80%',
            'height': '3em',
            'lineHeight': '3em',
            'borderWidth': '1px',
            'borderRadius': '5px',
            'borderStyle': 'dashed',
            'margin': '10px auto 0',
        },
    )

    check_icon = html.I(className='fa fa-circle-check')

    filename_field = dbc.Stack(
        [
            html.Div(id={'type': 'filename', 'index': identifier}),
            html.Div(check_icon, className='ms-auto'),
        ],
        gap=3,
        direction='horizontal',
        id={'type': 'filename-row', 'index': identifier},
        style={
            'width': '80%',
            'display': 'none',
            'fontSize': '0.75em',
            'margin': '5px auto 0',
        }
    )

    upload_block = [
        upload_label,
        horizontal_divider,
        file_uploader,
        filename_field,
    ]

    return upload_block


upload = html.Div(
    [
        *file_upload('Negative Electrode', 'neg'),
        *file_upload('Positive Electrode', 'pos'),
        *file_upload('Full Cell', 'cell'),
    ],
    id='upload-menu',
)


def number_input(id, low, high, step, value, **kwargs):
    input = dbc.Input(
        min=low,
        max=high,
        step=step,
        value=value,
        type='number',
        placeholder=value,
        id={'type': 'opt', 'index': id},
        **kwargs,
    )

    return dbc.Col(input, width=5, class_name='m-0')


def switch_input(id, value):
    switch = dbc.Switch(
        value=value,
        id={'type': 'cost', 'index': id},
    )

    return dbc.Col(switch, width=4, class_name='m-0')


opt_data = {
    'smoothing': 10,
    'xmin-bnd-neg': 0.1,
    'xmax-bnd-neg': 0.1,
    'xmin-bnd-pos': 0.1,
    'xmax-bnd-pos': 0.1,
    'coarse-Nx': 11,
    'max-iter': 1e5,
    'xtol': 1e-9,
    'voltage': True,
    'dqdv': True,
    'dvdq': True,
}

optimize = html.Div([
    dbc.Label('Model Parameters', class_name='bold-label'),
    html.Hr(className='m-0'),
    dbc.Row([
        dbc.Col(dbc.Label('Smoothing')),
        number_input('smoothing', 3, 25, 1,
                     opt_data['smoothing'], debounce=True),
    ], style={'width': '90%', 'margin': '5px auto'}),

    dbc.Label('Negative Electrode', class_name='bold-label'),
    html.Hr(className='m-0'),
    dbc.Row([
        dbc.Col(dbc.Label('xmin bounds (+/-)')),
        number_input('xmin-bnd-neg', 1e-3, 1, 1e-3, opt_data['xmin-bnd-neg']),
    ], style={'width': '90%', 'margin': '5px auto'}),
    dbc.Row([
        dbc.Col(dbc.Label('xmax bounds (+/-)')),
        number_input('xmax-bnd-neg', 1e-3, 1, 1e-3, opt_data['xmax-bnd-neg']),
    ], style={'width': '90%', 'margin': '5px auto'}),

    dbc.Label('Positive Electrode', class_name='bold-label'),
    html.Hr(className='m-0'),
    dbc.Row([
        dbc.Col(dbc.Label('xmin bounds (+/-)')),
        number_input('xmin-bnd-pos', 1e-3, 1, 1e-3, opt_data['xmin-bnd-pos']),
    ], style={'width': '90%', 'margin': '5px auto'}),
    dbc.Row([
        dbc.Col(dbc.Label('xmax bounds (+/-)')),
        number_input('xmax-bnd-pos', 1e-3, 1, 1e-3, opt_data['xmax-bnd-pos']),
    ], style={'width': '90%', 'margin': '5px auto'}),

    dbc.Label('Fitting Parameters', class_name='bold-label'),
    html.Hr(className='m-0'),
    dbc.Row([
        dbc.Col(dbc.Label('Coarse Nx')),
        number_input('coarse-Nx', 11, 51, 1, opt_data['coarse-Nx']),
    ], style={'width': '90%', 'margin': '5px auto'}),
    dbc.Row([
        dbc.Col(dbc.Label('Max Iterations')),
        number_input('max-iter', 1e3, 1e6, 1, opt_data['max-iter']),
    ], style={'width': '90%', 'margin': '5px auto'}),
    dbc.Row([
        dbc.Col(dbc.Label('x Tolerance')),
        number_input('xtol', 1e-15, 1e-2, 'any', opt_data['xtol']),
    ], style={'width': '90%', 'margin': '5px auto'}),

    dbc.Label('Cost Terms', class_name='bold-label'),
    html.Hr(className='m-0'),
    dbc.Row([
        dbc.Col(dbc.Label('Voltage')),
        switch_input('voltage', opt_data['voltage']),
    ], style={'width': '90%', 'margin': '5px auto'}),
    dbc.Row([
        dbc.Col(dbc.Label('Differential SOC')),
        switch_input('dqdv', opt_data['dqdv']),
    ], style={'width': '90%', 'margin': '5px auto'}),
    dbc.Row([
        dbc.Col(dbc.Label('Differential Voltage')),
        switch_input('dvdq', opt_data['dvdq']),
    ], style={'width': '90%', 'margin': '5px auto'}),
],
    id='optimize-menu',
)


def line_properties(label, identifier, values):
    linestyles = dict((v, v) for v in ['solid', 'dash', 'dot'])
    linewidths = dict((v, v) for v in [1, 2, 3, 4, 5])

    id_style = {'type': 'line-style', 'index': identifier}
    id_width = {'type': 'line-width', 'index': identifier}
    id_color = {'type': 'line-color', 'index': identifier}

    prop_label = dbc.Label(label, class_name='bold-label')

    horizontal_divider = html.Hr(className='m-0')

    linestyle = dbc.Row([
        dbc.Col(dbc.Label('Line style')),
        dbc.Col(dbc.Select(linestyles, value=values[0], id=id_style), width=5),
    ], style={'width': '90%', 'margin': '5px auto'})

    linewidth = dbc.Row([
        dbc.Col(dbc.Label('Line width')),
        dbc.Col(dbc.Select(linewidths, value=values[1], id=id_width), width=5),
    ], style={'width': '90%', 'margin': '5px auto'})

    linecolor = dbc.Row([
        dbc.Col(dbc.Label('Color')),
        dbc.Col(dbc.Input(type='color', value=values[2], id=id_color), width=5),
    ], style={'width': '90%', 'margin': '5px auto'})

    props_block = [
        prop_label,
        horizontal_divider,
        linestyle,
        linewidth,
        linecolor,
    ]

    return props_block


def marker_properties(label, identifier, values):
    markstyles = dict((v, v) for v in ['o', 'x', '^', '+'])
    marksizes = dict((v, v) for v in [5, 10, 15, 20, 25])

    id_style = {'type': 'mark-style', 'index': identifier}
    id_size = {'type': 'mark-size', 'index': identifier}
    id_color = {'type': 'mark-color', 'index': identifier}

    prop_label = dbc.Label(label, class_name='bold-label')

    horizontal_divider = html.Hr(className='m-0')

    linestyle = dbc.Row([
        dbc.Col(dbc.Label('Marker style')),
        dbc.Col(dbc.Select(markstyles, value=values[0], id=id_style), width=5),
    ], style={'width': '90%', 'margin': '5px auto'})

    linewidth = dbc.Row([
        dbc.Col(dbc.Label('Line width')),
        dbc.Col(dbc.Select(marksizes, value=values[1], id=id_size), width=5),
    ], style={'width': '90%', 'margin': '5px auto'})

    linecolor = dbc.Row([
        dbc.Col(dbc.Label('Color')),
        dbc.Col(dbc.Input(type='color', value=values[2], id=id_color), width=5),
    ], style={'width': '90%', 'margin': '5px auto'})

    props_block = [
        prop_label,
        horizontal_divider,
        linestyle,
        linewidth,
        linecolor,
    ]

    return props_block


neg_style = ['solid', 2, '#dc3912']
pos_style = ['solid', 2, '#3366cc']
cell_style = ['o', 10, '#990099']
model_style = ['solid', 2, '#999999']

fig_menu = html.Div(
    [
        *line_properties('Negative Electrode', 'neg', neg_style),
        *line_properties('Positive Electrode', 'pos', pos_style),
        *marker_properties('Full Cell', 'cell', cell_style),
        *line_properties('Model Fits', 'model', model_style),
    ],
    id='figure-menu',
)

up_icon = html.Div(
    [
        html.I(className='fa fa-file-import sidebar-icon'),
        html.P('Upload Data', className='m-0'),
    ],
    className='d-flex align-items-center',
)

opt_icon = html.Div(
    [
        html.I(className='fa fa-gears sidebar-icon'),
        html.P('Optimization Settings', className='m-0'),
    ],
    className='d-flex align-items-center',
)

fig_icon = html.Div(
    [
        html.I(className='fa fa-chart-line sidebar-icon'),
        html.P('Figure Options', className='m-0'),
    ],
    className='d-flex align-items-center',
)

accordion = dbc.Accordion(
    [
        dbc.AccordionItem(upload, title=up_icon, class_name='sidebar'),
        dbc.AccordionItem(optimize, title=opt_icon, class_name='sidebar'),
        dbc.AccordionItem(fig_menu, title=fig_icon, class_name='sidebar'),
    ],
    flush=True,
    class_name='accordion',
    style={'width': '100%'},
)

opt_store = dcc.Store(
    data=opt_data,
    id='opt-store',
)

flags = dcc.Store(
    id='flags',
    data={'neg': False, 'pos': False, 'cell': False},
)

summary_store = dcc.Store(
    data={},
    id='summary-store',
)

sidebar = dbc.Offcanvas(
    id='sidebar',
    scrollable=True,
    class_name='sidebar',
    children=[opt_store, flags, summary_store, accordion],
)

# Toggle sidebar for small screens
dash.clientside_callback(
    """
    function toggleSidebar(nClick, isOpen) {
        return !isOpen;
    }
    """,
    Output('sidebar', 'is_open', allow_duplicate=True),
    Input('sidebar-btn', 'n_clicks'),
    State('sidebar', 'is_open'),
    prevent_initial_call=True,
)

# Reset optimization values to placeholder on blur if invalid
dash.clientside_callback(
    """
    function(blur, value, min, max, placeholder) {
        if (value >= min && value <= max) {
            return value;
        }

        return placeholder;
    }
    """,
    Output({'type': 'opt', 'index': dash.MATCH}, 'value'),
    Input({'type': 'opt', 'index': dash.MATCH}, 'n_blur'),
    State({'type': 'opt', 'index': dash.MATCH}, 'value'),
    State({'type': 'opt', 'index': dash.MATCH}, 'min'),
    State({'type': 'opt', 'index': dash.MATCH}, 'max'),
    State({'type': 'opt', 'index': dash.MATCH}, 'placeholder'),
    prevent_initial_call=True,
)

# Sync optimization input changes to dcc.Store
dash.clientside_callback(
    """
    function(values, data) {
        const triggered = dash_clientside.callback_context.triggered;
        const id = triggered[0].prop_id.split(".")[0];
        const idx = JSON.parse(id).index;

        data[idx] = triggered[0].value;

        return data;
    }
    """,
    Output('opt-store', 'data', allow_duplicate=True),
    Input({'type': 'opt', 'index': dash.ALL}, 'value'),
    State('opt-store', 'data'),
    prevent_initial_call=True,
)

# Ensure at least one cost switch is always active
dash.clientside_callback(
    """
    function(switchValues, data) {
        const triggered = dash_clientside.callback_context.triggered;
        const id = triggered[0].prop_id.split(".")[0];
        const idx = JSON.parse(id).index;

        data[idx] = triggered[0].value;

        const enabledCount = switchValues.filter(v => v).length;
        const switchStates = switchValues.map(v => enabledCount === 1 && v);

        return [data, switchStates];
    }
    """,
    Output('opt-store', 'data', allow_duplicate=True),
    Output({'type': 'cost', 'index': dash.ALL}, 'disabled'),
    Input({'type': 'cost', 'index': dash.ALL}, 'value'),
    State('opt-store', 'data'),
    prevent_initial_call=True,
)

# Support functions


def make_figure(params, flags, new_data=False):

    if new_data and all(flags.values()):
        output = fitter.err_terms(params, full_output=True)

        x = output['soc']
        xm = output['soc_mid']

        y1d = output['V_dat']
        y2d = output['dqdv_dat']
        y3d = output['dvdq_dat']

        y1f = output['V_fit']
        y2f = output['dqdv_fit']
        y3f = output['dvdq_fit']

        figure.data = ()

        mk = {}  # dict(color='#990099', size=10, symbol='circle')
        ln = {}  # dict(color='#999999', width=2, dash='solid')

        dat = dict(mode='markers', name='Full Cell',
                   showlegend=False, marker=mk)
        fit = dict(mode='lines', name='Model', showlegend=False, line=ln)

        figure.add_trace(go.Scatter(x=x, y=y1d, **dat), row=1, col=1)
        figure.add_trace(go.Scatter(x=xm, y=y2d, **dat), row=1, col=2)
        figure.add_trace(go.Scatter(x=xm, y=y3d, **dat), row=1, col=3)

        figure.add_trace(go.Scatter(x=x, y=y1f, **fit), row=1, col=1)
        figure.add_trace(go.Scatter(x=xm, y=y2f, **fit), row=1, col=2)
        figure.add_trace(go.Scatter(x=xm, y=y3f, **fit), row=1, col=3)

    elif all(flags.values()):
        output = fitter.err_terms(params, full_output=True)

        figure.data[3].y = output['V_fit']
        figure.data[4].y = output['dqdv_fit']
        figure.data[5].y = output['dvdq_fit']

    if all(flags.values()):
        return figure
    else:
        return placeholder_fig


UPLOAD_IDS = ['neg', 'pos', 'cell']


@dash.callback(
    Output('figure-div', 'figure', allow_duplicate=True),
    Output('flags', 'data', allow_duplicate=True),
    Input({'type': 'upload', 'index': dash.ALL}, 'contents'),
    State('neg-slider', 'value'),
    State('pos-slider', 'value'),
    State('flags', 'data'),
    prevent_initial_call=True,
)
def upload_data(contents_list, neg_s, pos_s, flags):
    trigger = dash.callback_context.triggered[0]['prop_id'].split('.')[0]
    key = eval(trigger)['index']

    contents = contents_list[UPLOAD_IDS.index(key)]

    if contents is not None:
        _, content_string = contents.split(',')
        decoded = base64.b64decode(content_string)

        setattr(fitter, 'df_' + key, pd.read_csv(io.BytesIO(decoded)))
        flags[key] = True

    params = np.array(neg_s + pos_s)

    figure = make_figure(params, flags, new_data=True)

    return figure, flags


@dash.callback(
    Output({'type': 'filename', 'index': dash.MATCH}, 'children'),
    Output({'type': 'filename-row', 'index': dash.MATCH}, 'style'),
    Input({'type': 'upload', 'index': dash.MATCH}, 'filename'),
    State({'type': 'filename-row', 'index': dash.MATCH}, 'style'),
    prevent_initial_call=True,
)
def show_filename(filename, style):
    new_style = {**style, 'display': 'flex'}
    return f"{filename}", new_style


@dash.callback(
    Output('figure-div', 'figure', allow_duplicate=True),
    Input({'type': 'opt', 'index': 'smoothing'}, 'value'),
    State('neg-slider', 'value'),
    State('pos-slider', 'value'),
    State('flags', 'data'),
    prevent_initial_call=True,
)
def update_on_smoothing(smoothing, neg_s, pos_s, flags):
    fitter.smoothing = smoothing
    params = np.array(neg_s + pos_s)
    figure = make_figure(params, flags)
    return figure


@dash.callback(
    Output('figure-div', 'figure', allow_duplicate=True),
    Output('neg-slider-label', 'children', allow_duplicate=True),
    Output('pos-slider-label', 'children', allow_duplicate=True),
    Input('neg-slider', 'value'),
    Input('pos-slider', 'value'),
    State('flags', 'data'),
    prevent_initial_call=True,
)
def update_on_slider(neg_s, pos_s, flags):
    global optimal_params

    params = np.array(neg_s + pos_s)
    if isinstance(optimal_params, np.ndarray):
        figure = make_figure(optimal_params, flags)
    else:
        figure = make_figure(params, flags)

    neg_label = f"Negative Electrode: [{neg_s[0]:.2f}, {neg_s[1]:.2f}]"
    pos_label = f"Positive Electrode: [{pos_s[0]:.2f}, {pos_s[1]:.2f}]"

    optimal_params = False

    return figure, neg_label, pos_label


@dash.callback(
    Output('spinner-div', 'children', allow_duplicate=True),
    Output('neg-slider', 'value', allow_duplicate=True),
    Output('pos-slider', 'value', allow_duplicate=True),
    Output('summary-store', 'data'),
    Input('coarse-btn', 'n_clicks'),
    Input('min-err-btn', 'n_clicks'),
    State('neg-slider', 'value'),
    State('pos-slider', 'value'),
    State('opt-store', 'data'),
    State('flags', 'data'),
    prevent_initial_call=True,
)
def update_on_button(_c, _m, neg_s, pos_s, opt_data, flags):
    global optimal_params

    trigger = dash.callback_context.triggered[0]['prop_id'].split('.')[0]

    fitter.xtol = opt_data['xtol']
    fitter.maxiter = opt_data['max-iter']
    fitter.smoothing = opt_data['smoothing']
    fitter.bounds = [
        opt_data['xmin-bnd-neg'],
        opt_data['xmax-bnd-neg'],
        opt_data['xmin-bnd-pos'],
        opt_data['xmax-bnd-pos'],
    ]

    cost_terms = []
    if opt_data['voltage']:
        cost_terms.append('voltage')
    if opt_data['dqdv']:
        cost_terms.append('dqdv')
    if opt_data['dvdq']:
        cost_terms.append('dvdq')

    fitter.cost_terms = cost_terms

    summary = {}
    if trigger == 'coarse-btn' and _c and all(flags.values()):
        summary = fitter.coarse_search(opt_data['coarse-Nx'])
        params = summary['x']
    elif trigger == 'min-err-btn' and all(flags.values()):
        x0 = np.array(neg_s + pos_s)
        summary = fitter.constrained_fit(x0)
        params = summary['x']
    else:
        params = np.array(neg_s + pos_s)

    optimal_params = params.copy()

    if summary:
        neg_s[0] = round(summary['x'][0], 2)
        neg_s[1] = round(summary['x'][1], 2)

        pos_s[0] = round(summary['x'][2], 2)
        pos_s[1] = round(summary['x'][3], 2)

    return '', neg_s, pos_s, summary


@dash.callback(
    Output('figure-div', 'figure', allow_duplicate=True),
    Input('theme-switch', 'value'),
    State('flags', 'data'),
    prevent_initial_call=True,
)
def toggle_theme_switch(switch_on, flags):
    if switch_on:
        border_color = '#212529'
        template = pio.templates['bootstrap']
    else:
        border_color = '#DEE2E6'
        template = pio.templates['bootstrap_dark']

    figure.update_layout(template=template)

    for i in range(3):
        figure.update_xaxes(
            row=1, col=i+1,
            linecolor=border_color, tickcolor=border_color,
        )
        figure.update_yaxes(
            row=1, col=i+1,
            linecolor=border_color, tickcolor=border_color,
        )

    if all(flags.values()):
        return figure
    else:
        return placeholder_fig



"""
TODO
----

"""

from numpy import ndarray as _ndarray


def reset_rcparams() -> None:
    """
    Sets ``plt.rcParams`` back to defaults.

    Returns
    -------
    None.

    """

    import matplotlib.pyplot as plt

    plt.rcdefaults()


def set_tick_rcparams(allsides: bool = True, minorticks: bool = True,
                      direction: str = 'in') -> None:
    """
    Sets ``plt.rcParams`` tick details.

    Parameters
    ----------
    allsides : bool, optional
        Turns on ticks for top and right sides. The default is True.
    minorticks : bool, optional
        Makes minor ticks visible. The default is True.
    direction : str, optional
        Tick direction from {'in', 'out'}. The default is 'in'.

    Returns
    -------
    None.

    """

    import matplotlib.pyplot as plt

    if allsides:
        plt.rcParams['xtick.top'] = True
        plt.rcParams['ytick.right'] = True

    if minorticks:
        plt.rcParams['xtick.minor.visible'] = True
        plt.rcParams['ytick.minor.visible'] = True

    if direction not in ['in', 'out']:
        raise ValueError(f"{direction=} is not a valid value; supported"
                         + " values are {'in', 'out'}.")
    else:
        plt.rcParams['xtick.direction'] = direction
        plt.rcParams['ytick.direction'] = direction


def set_font_rcparams(fontsize: int = 10, family: str = 'sans-serif') -> None:
    """
    Sets ``plt.rcParams`` font details.

    Parameters
    ----------
    fontsize : int, optional
        Font size to use across all figures. The default is 10.
    family : str, optional
        Font family from {'serif', 'sans-serif'}. The default is 'sans-serif'.

    Returns
    -------
    None.

    """

    import matplotlib.pyplot as plt

    plt.rcParams['font.size'] = fontsize

    if family == 'serif':
        plt.rcParams['font.family'] = 'serif'
        plt.rcParams['mathtext.fontset'] = 'dejavuserif'
    elif family == 'sans-serif':
        plt.rcParams['font.family'] = 'sans-serif'
        plt.rcParams['mathtext.fontset'] = 'dejavusans'
    else:
        raise ValueError(f"{family=} is not a valid value; supported"
                         + " values are {'serif', 'sans-serif'}.")


def format_ticks(ax: object, xdiv: int = None, ydiv: int = None) -> None:
    """
    Formats ``ax`` ticks.

    Top and right ticks are turned on, tick direction is set to 'in', and
    minor ticks are made visible with the specified number of subdivisions.

    Parameters
    ----------
    ax : object
        An ``axis`` instance from a ``matplotlib`` figure.
    xdiv : int, optional
        Number of divisions between x major ticks. The default is None, which
        performs an 'auto' subdivision.
    ydiv : int, optional
        Number of divisions between y major ticks. The default is None, which
        performs an 'auto' subdivision.

    Returns
    -------
    None.

    """

    import matplotlib as mpl

    if ax.get_xaxis().get_scale() != 'log':
        ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(xdiv))

    if ax.get_yaxis().get_scale() != 'log':
        ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(ydiv))

    ax.tick_params(axis='x', top=True, which='both', direction='in')
    ax.tick_params(axis='y', right=True, which='both', direction='in')


def add_text(ax: object, xloc: float, yloc: float, text: str) -> None:
    """
    Adds text to ``ax`` at a specified location.

    Parameters
    ----------
    ax : object
        An ``axis`` instance from a ``matplotlib`` figure.
    xloc : float
        Relative location (0-1) for text start in x-direction.
    yloc : float
        Relative location (0-1) for text start in y-direction.
    text : str
        Text string to add to figure.

    Returns
    -------
    None.

    """

    ax.text(xloc, yloc, text, transform=ax.transAxes)


def cb_line_plot(ax: object, xdata: list[_ndarray], ydata: list[_ndarray],
                 zdata: _ndarray, cmap: str = 'jet', **kwargs) -> None:
    """
    TODO

    Parameters
    ----------
    ax : object
        _description_
    xdata : list[_ndarray]
        _description_
    ydata : list[_ndarray]
        _description_
    zdata : _ndarray
        _description_
    cmap : str, optional
        _description_, by default 'jet'

    Raises
    ------
    ValueError
        _description_

    """

    import numpy as np
    import matplotlib as mpl

    cmap = mpl.colormaps[cmap]

    cbtype = kwargs.pop('cbtype', 'continuous')
    if cbtype == 'continuous':
        zmin = kwargs.pop('zmin', zdata.min())
        zmax = kwargs.pop('zmax', zdata.max())

        norm = mpl.colors.Normalize(vmin=zmin, vmax=zmax)

        cbticks = kwargs.pop('cbticks', None)
        cbticklabels = kwargs.pop('cbticklabels', None)

    elif cbtype == 'discrete':
        zmin = kwargs.pop('zmin', 0)
        zmax = kwargs.pop('zmax', zdata.size + 1)

        norm = mpl.colors.BoundaryNorm(np.arange(zmin, zmax, 1), cmap.N)

        cbticks = kwargs.pop('cbticks', np.arange(zmin + 0.5, zmax + 0.5, 1))
        cbticklabels = kwargs.pop('cbticklabels', zdata)

    else:
        raise ValueError(f"{cbtype=} is not a valid value; supported"
                         + " values are {'continuous', 'discrete'}.")

    sm = mpl.pyplot.cm.ScalarMappable(cmap=cmap, norm=norm)
    if cbtype == 'continuous':
        colors = [sm.to_rgba(z) for z in zdata]
    elif cbtype == 'discrete':
        colors = [sm.to_rgba(z) for z in range(zdata.size + 1)]

    ls = kwargs.pop('linestyle', '-')
    for x, y, color in zip(xdata, ydata, colors):
        ax.plot(x, y, linestyle=ls, color=color)

    cax = kwargs.pop('cax', ax)
    if cax is not None:
        cb = mpl.pyplot.colorbar(sm, ax=cax, ticks=cbticks)

        if cbticklabels is not None:
            cb.set_ticklabels(cbticklabels)

        if cbtype == 'discrete':
            cb.minorticks_off()

        cblabel = kwargs.pop('cblabel', None)
        if cblabel is not None:
            cb.set_label(cblabel)

    if kwargs.pop('format_ticks', True):
        format_ticks(ax, xdiv=kwargs.get('xdiv', None),
                     ydiv=kwargs.get('ydiv', None))

    if len(kwargs) != 0:
        ax.set(**kwargs)



"""
TODO
----

"""


def alphanumeric_sort(unsorted_list: list[str]) -> list[str]:
    """
    TODO

    Parameters
    ----------
    unsorted_list : list[str]
        _description_

    Returns
    -------
    list[str]
        _description_

    """
    import re

    def convert(txt): return int(txt) if txt.isdigit() else txt
    def alphanum(key): return [convert(c) for c in re.split('([0-9]+)', key)]

    sorted_list = sorted(unsorted_list, key=alphanum)

    return sorted_list


class ProgressBar:
    """Progress bar"""

    __slots__ = ['width']

    def __init__(self, width: int = 50) -> None:
        """
        TODO

        Parameters
        ----------
        width : int, optional
            _description_, by default 50

        """
        self.width = width

    def update(self, percent) -> None:
        """
        TODO

        Parameters
        ----------
        percent : _type_
            _description_

        """
        import sys

        done = int(percent * self.width)
        wait = self.width - int(percent * self.width)

        bar = '[' + '#' * done + '-' * wait + ']'
        sys.stdout.write('\r' + f'Progress: {bar} {percent * 100:.2f}%')
        sys.stdout.flush()



import dash_bootstrap_components as dbc
from dash import html, Output, Input, State, clientside_callback

# Branding
PLOTLY_LOGO = 'https://images.plot.ly/logo/new-branding/plotly-logomark.png'

brand_logo = html.Div(
    [
        html.Img(src=PLOTLY_LOGO, height='30px'),
        html.P(children='Logo', className='mx-3 my-auto',
               style={'fontSize': '1.75em', 'fontWeight': 'bold'}),
    ],
    style={'height': '4em'},
    className='d-flex align-items-center',
)

# Buttons/switches
sidebar_btn = dbc.Button(
    id='sidebar-btn',
    class_name='nav-btn ms-3 me-2',
    children=html.I(className='fa fa-bars'),
)

theme_btn = dbc.Button(
    id='theme-btn',
    class_name='nav-btn',
    children=html.I(id='theme-btn-icon', className='fa fa-sun'),
)

theme_switch = html.Div(
    [
        html.I(className='fa fa-lg fa-moon'),
        dbc.Switch(id='theme-switch', value=True, class_name='ms-2'),
        html.I(className='fa fa-lg fa-sun'),
    ],
    className='d-flex justify-content-end align-items-center',
)

navbar_btn = dbc.Button(
    id='navbar-btn',
    class_name='nav-btn me-3',
    children=html.I(className='fa fa-ellipsis-vertical'),
)

# Page links
user_guide = dbc.NavLink(
    href='/',
    active='exact',
    children='User Guide',
)

dvq_analysis = dbc.NavLink(
    active='exact',
    href='/dqv-analysis',
    children='DVQ Analysis',
    class_name='nav-link-bottom',
)

navlinks = dbc.Nav(
    [
        user_guide,
        dvq_analysis,
    ],
)

# Full navbar
navbar = dbc.Navbar(
    dbc.Container(
        [
            dbc.Stack(
                className='sm-only',
                direction='horizontal',
                children=[sidebar_btn, brand_logo],
            ),
            html.Div([brand_logo], className='lg-only ps-4 w-400px'),
            html.Div([theme_btn, navbar_btn], className='sm-only'),
            dbc.Collapse(
                navbar=True,
                is_open=False,
                children=navlinks,
                id='navlinks-collapse',
                class_name='navlinks-collapse mx-2',
            ),
            html.Div([theme_switch], className='lg-only ms-auto pe-4'),
        ],
        fluid=True,
        class_name='p-0 d-flex align-items-center',
    ),
    id='navbar',
    fixed='top',
    expand='xl',
    class_name='p-0 header',
)

# Toggle navbar when small
clientside_callback(
    """
    function toggleNavCollapse(nClick, isOpen) {
        return !isOpen;
    }
    """,
    Output('navlinks-collapse', 'is_open'),
    Input('navbar-btn', 'n_clicks'),
    State('navlinks-collapse', 'is_open'),
    prevent_initial_call=True,
)

# Change theme - switch (light vs. dark)
clientside_callback(
    """
    function themeSwitch(switchOn) {
        document.documentElement.setAttribute(
            "data-bs-theme", switchOn ? "light" : "dark",
        );
        return window.dash_clientside.no_update;
    }
    """,
    Output('theme-switch', 'id'),
    Input('theme-switch', 'value'),
)

# Change theme - button (light vs. dark)
clientside_callback(
    """
    function themeButton(click, themeSwitch) {
        return !themeSwitch;
    }
    """,
    Output('theme-switch', 'value'),
    Input('theme-btn', 'n_clicks'),
    State('theme-switch', 'value'),
    prevent_initial_call=True,
)



from __future__ import annotations

from typing import TYPE_CHECKING
from dataclasses import dataclass

import numpy as np
import matplotlib.pyplot as plt

if TYPE_CHECKING:  # pragma: no cover
    import numpy.typing as npt


@dataclass(slots=True)
class CellDescription:
    """
    Cell description wrapper

    A dataclass to describe the cell. Requires cell geometry, microstructure,
    and material properties information. Make sure parameters are input with
    the correct units, as described in the documentation.

    Parameters
    ----------
    thick_ed : float
        Electrode thickness [m].
    area_ed : float
        Projected electrode area (e.g., pi*R**2 for coin cells, L*W for pouch
        cells, etc.) [m2].
    eps_el : float
        Electrolyte/pore volume fraction [-].
    eps_CBD : float
        Carbon-binder-domain volume fraction [-]. See notes for more info.
    radius_AM : float
        Active material particle radius [m].
    rho_AM : float
        Active material mass density [kg/m3].
    mass_AM : float
        Total active material mass [kg].
    molar_mass_AM : float
        Active material molar mass [kg/kmol].

    Returns
    -------
    None.

    Notes
    -----
    A "convenient" way to get ``eps_CBD`` requires knowledge of the densities
    and masses for all solid phases in your slurry (carbon additive, binder,
    and active material). The volume fraction for any phase :math:`m` is

    .. math::

        \\varepsilon_{m} = f_{m} \\varepsilon_{\\rm s},

    where :math:`f_{m}` is the volume of phase :math:`m` per volume of solid
    phase and :math:`\\varepsilon_{\\rm s} = 1 - \\varepsilon_{\\rm el}` is
    the total solid-phase volume fraction. :math:`f_{m}` is calculated as

    .. math::

        f_{m} = \\frac{m_{m} / \\rho_{m}}{\\sum_{i=1}^{N} m_{i} / \\rho_{i}},

    Here, the numerator uses the mass and density of phase :math:`m` to get
    its individual volume, and the denominator sums over all :math:`N` solid
    phases to calculate the total solid-phase volume. Using these expressions,
    you can separately calculate volume fractions for the carbon additive and
    binder. Finally, adding their values together gives

    .. math::

        \\varepsilon_{\\rm CBD} = \\varepsilon_{\\rm C}
                                + \\varepsilon_{\\rm B}.

    """

    thick_ed: float
    area_ed: float
    eps_el: float
    eps_CBD: float
    radius_AM: float
    rho_AM: float
    mass_AM: float
    molar_mass_AM: float

    @property
    def volume_ed(self) -> float:
        """Electrode volume [m3]."""
        return self.thick_ed*self.area_ed

    @property
    def spec_capacity_AM(self) -> float:
        """Active material theoretical specific capacity [Ah/kg]."""
        return 96485.33e3 / (3600.*self.molar_mass_AM)

    @property
    def molar_vol_AM(self) -> float:
        """Active material molar volume [m3/kmol]."""
        return self.molar_mass_AM / self.rho_AM

    @property
    def eps_AM(self) -> float:
        """Active material volume fraction [-]."""
        return 1. - self.eps_el - self.eps_CBD

    @property
    def surf_area_AM(self) -> float:
        """Total active material surface area [m2]."""
        return 3.*self.eps_AM*self.volume_ed / self.radius_AM


class GITTDataset:
    """GITT dataclass wrapper"""

    def __init__(self, time: npt.ArrayLike, current: npt.ArrayLike,
                 voltage: npt.ArrayLike, avg_temperature: float,
                 invert_current: bool = False) -> None:
        """
        A dataclass to wrap experimental GITT data.

        Parameters
        ----------
        time : ArrayLike, shape(n,)
            Recorded test times [s].
        current : ArrayLike, shape(n,)
            Timeseries current data [A].
        voltage : ArrayLike, shape(n,)
            Timeseries voltage data [V].
        avg_temperature : float
            Average temperature of the experiment [K].
        invert_current : bool, optional
            Inverts the 'current' sign values. Charge and discharge currents
            should be positive and negative, respectively. Defaults to False.

        Returns
        -------
        None.

        Raises
        ------
        ValueError
            'time' array must be increasing.

        """

        self.time = np.asarray(time)
        self.current = np.asarray(current)
        self.voltage = np.asarray(voltage)
        self.avg_temperature = avg_temperature

        if not all(np.diff(self.time) >= 0.):
            raise ValueError("'time' array must be increasing.")

        if invert_current:
            self.current *= -1.

    def find_pulses(self, pulse_sign: int, plot: bool = False) -> tuple[int]:
        """
        Finds the indices in the data where pulses start and end. The algorithm
        depends on there being a rest period both before and after each pulse.

        Parameters
        ----------
        pulse_sign : int
            The sign of the current pulses to find. Use `+1` or `-1` for
            positive and negative pulses, respectively.
        plot : bool, optional
            Whether or not to plot the result. The default is False.

        Returns
        -------
        start : int
            Indices where pulse starts were detected.
        stop : int
            Indices where pulse stops were detected.

        Raises
        ------
        ValueError
            Invalid pulse_sign value, must be +1 or -1.
        ValueError
            Size mismatch: The number of detected pulse starts and stops do
            not agree. This typically occurs due to a missing rest. You will
            likely need to manually remove affected pulse(s).

        """

        I_pulse = self._avg_pulse_current(pulse_sign)

        if pulse_sign == 1:
            I_tmp = np.where(self.current > 0.5*I_pulse, 1, 0)
        elif pulse_sign == -1:
            I_tmp = np.where(self.current < 0.5*I_pulse, 1, 0)

        idx1 = np.where(np.diff(I_tmp) > 0.9)[0]
        idx2 = np.where(I_tmp > -0.9)[0]
        start = np.intersect1d(idx1, idx2)

        idx1 = np.where(I_tmp > 0.9)[0]
        idx2 = np.where(np.diff(I_tmp) < -0.9)
        stop = np.intersect1d(idx1, idx2)

        if plot:
            plt.figure()
            plt.plot(self.time / 3600., 1e3*self.current, '-k')
            plt.plot(self.time[start] / 3600., 1e3*self.current[start], 'sg')
            plt.plot(self.time[stop] / 3600., 1e3*self.current[stop], 'or')

            plt.xlabel('Time [h]')
            plt.ylabel('Current [mA]')

            plt.show()

        return start, stop

    def get_stats(self, pulse_sign: int) -> dict:

        time = self.time.copy()
        I_pulse = self._avg_pulse_current(pulse_sign)
        start, stop = self.find_pulses(pulse_sign, plot=False)

        stats = {
            'num pulses': start.size,
            'avg I_pulse [A]': I_pulse,
            'avg t_pulse [s]': np.mean(time[stop] - time[start]),
            'avg t_rest [s]': np.mean(time[start[1:]] - time[stop[:-1]]),
        }

        return stats

    def plot(self, y_key: str, t_units: str = 'h') -> None:

        plt.figure()
        if y_key == 'current':
            y, ylabel = self.current.copy(), 'Current [A]'
        elif y_key == 'voltage':
            y, ylabel = self.voltage.copy(), 'Voltage [V]'
        else:
            raise ValueError()

        converter = {
            's': lambda t: t,
            'min': lambda t: t / 60.,
            'h': lambda t: t / 3600.,
            'day': lambda t: t / 3600. / 24.,
        }

        x, xlabel = converter[t_units](self.time), f"Time [{t_units}]"

        plt.plot(x, y)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        plt.show()

    def _avg_pulse_current(self, pulse_sign: int) -> float:

        if pulse_sign == 1:
            I_pulse = np.mean(self.current[self.current > 0.])
        elif pulse_sign == -1:
            I_pulse = np.mean(self.current[self.current < 0.])
        else:
            raise ValueError(f"Invalid {pulse_sign=}, must be +1 or -1.")

        return I_pulse



import argparse


def main():
    parser = argparse.ArgumentParser(description='CLI for ampworks')

    parser.add_argument(
        '--app',
        required=True,
        choices=['dQdV'],
        help='name of app to open',
    )

    parser.add_argument(
        '--debug',
        action='store_true',
        help='enables debug mode',
    )

    args = parser.parse_args()

    if args.app.lower() == 'dqdv':
        from .dqdv.gui_files._gui import run
        run(args.debug)
    else:
        print("No valid argument provided.")



"""
Summary
=======
A toolkit for battery data analysis.

Accessing the Documentation
---------------------------
Documentation is accessible via Python's ``help()`` function which prints
docstrings from a package, module, function, class, etc. You can also access
the documentation by visiting the website, hosted on Read the Docs. The website
includes search functionality and more detailed examples.

"""

__version__ = '0.0.1'

__all__ = [
    'dqdv',
    'gitt',
    'plotutils',
    'utils',
]


def __getattr__(attr):

    if attr == 'dqdv':
        import ampworks.dqdv as dqdv
        return dqdv
    elif attr == 'gitt':
        import ampworks.gitt as gitt
        return gitt
    elif attr == 'plotutils':
        import ampworks.plotutils as plotutils
        return plotutils
    elif attr == 'utils':
        import ampworks.utils as utils
        return utils


def __dir__():
    public_symbols = (globals().keys() | __all__)
    return list(public_symbols)



from dash import dcc
import plotly.graph_objs as go
from plotly.subplots import make_subplots

placeholder_fig = go.Figure()

placeholder_fig.update_layout(
    uirevision='constant',
    plot_bgcolor='lightgrey',
    xaxis=dict(visible=False),
    yaxis=dict(visible=False),
    paper_bgcolor='rgba(0, 0, 0, 0)',
    margin=dict(l=20, r=20, t=20, b=20),
)

placeholder_fig.add_annotation(
    align='center', showarrow=False,
    font=dict(size=20, color='black'),
    text='Upload data to populate figure',
    xref='paper', yref='paper', x=0.5, y=0.5,
)

figure = make_subplots(1, 3)

figure.update_layout(
    uirevision='constant',
    margin=dict(l=20, r=20, t=20, b=60),
)

y_labels = ['Voltage [V]', 'dsoc/dV [1/V]', 'dV/dsoc [V]']

for i in range(3):
    figure.update_xaxes(
        row=1, col=i+1,
        ticks='inside', tickcolor='#212529',
        mirror='allticks', title_text='SOC [-]',
        showline=True, linewidth=1, linecolor='#212529',
    )
    figure.update_yaxes(
        row=1, col=i+1,
        ticks='inside', tickcolor='#212529',
        mirror='allticks', title_text=y_labels[i],
        showline=True, linewidth=1, linecolor='#212529',
    )

figure_div = dcc.Graph(
    id='figure-div',
    responsive=True,
    figure=placeholder_fig,
    style={'height': '370px', 'width': '100%'},
    config={
        'scrollZoom': True,
        'displaylogo': False,
        'modeBarButtonsToRemove': ['lasso2d', 'select2d'],
    },
)



from numpy import ndarray as _ndarray
from scipy.interpolate import Akima1DInterpolator as _akima


def dydx(x: _ndarray, y: _ndarray, **kwargs) -> _ndarray:
    """
    Calculate numerical derivative dy/dx.

    Parameters
    ----------
    x : 1D array
        Independent variable.
    y : 1D array
        Dependent variable.
    **kwargs : dict, optional
        Keyword arguments. Optional key/value pairs and defaults given below.

        ================= =============================================
        Key               Value (*type*, default)
        ================= =============================================
        smoothing_win_x   smoothing window for x data (*float*, None)
        smoothing_win_y   smoothing window for y data (*float*, None)
        lead_value        leading derivative value (*float*, None)
        ================= =============================================

    Returns
    -------
    dy_dx : ndarray
        First-order numerical derivative dy/dx.

    """

    import numpy as np
    import scipy.signal as sig

    sw_x = kwargs.get('smoothing_win_x', None)
    sw_y = kwargs.get('smoothing_win_y', None)
    lead_value = kwargs.get('lead_value', None)

    if sw_x is not None:
        x = sig.savgol_filter(x, sw_x, 2)

    if sw_y is not None:
        y = sig.savgol_filter(y, sw_y, 2)

    dy_dx = np.diff(y) / np.diff(x)
    if lead_value is not None:
        dy_dx = np.hstack([lead_value, dy_dx])

    return dy_dx


def combinations(values: list[_ndarray], names: list[str] = []) -> list[dict]:
    """
    Generate all value combinations.

    Parameters
    ----------
    values : list[1D array]
        Possible variable values. The array with index ``i`` corresponds to
        the variable ``names[i]``, if provided.
    names : list[str], optional
        Variable names. If not provided, the dictionary will use integers
        from ``0`` to ``len(names) - 1`` as the keys for the output.

    Returns
    -------
    combinations : list[dict]
        Dictionaries for each possible combination of values.

    """

    import itertools

    if names == []:
        names = [i for i in range(len(values))]

    combinations = []
    for combination in itertools.product(*values):
        combinations.append({k: v for k, v in zip(names, combination)})

    return combinations


class makima(_akima):

    def __init__(self, x: _ndarray, y: _ndarray,
                 extrapolate: bool = True) -> None:
        """
        Modified Akima interpolator.

        Parameters
        ----------
        x : 1D array
            Independent variable data.
        y : 1D array
            Dependent variable data.

        """

        super().__init__(x, y)
        self.extrapolate = extrapolate



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from scipy.stats import linregress
from scipy.integrate import trapezoid
from scipy.interpolate import Akima1DInterpolator

if TYPE_CHECKING:  # pragma: no cover
    from ._dataclasses import CellDescription, GITTDataset


def extract_params(pulse_sign: int, cell: CellDescription, data: GITTDataset,
                   return_stats: bool = False, **options) -> dict:
    """_summary_

    Parameters
    ----------
    pulse_sign : int
        The sign of the current pulses to process. Use `+1` for positive pulses
        and `-1` for negative pulses.
    cell : CellDescription
        Description of the cell.
    data : GITTDataset
        The GITT data to process.
    return_stats : bool, optional
        Adds a second return value with some statistics from the experiment,
        see below. The default is False.
    **options : dict, optional
        Keyword options to further control the function behavior. A full list
        of names, types, descriptions, and defaults is given below.
    xs_ref : float, optional
        Shifts the intercalation fraction output value such that the lower or
        upper bound (set via `ref_location`) is `xs_ref`. The default is 1.
    ref_location : {'lower', 'upper'}, optional
        Specifies whether the `xs_ref` value is used as a lower or upper bound.
        The default is 'upper'.
    R2_lim : float, optional
        Lower limit for the coefficient of determination. Pulses whose linear
        regression for `sqrt(time)` vs `voltage` that are less than this value
        result in a diffusivity of `nan`. The default is 0.95.
    replace_nans : bool, optional
        If True (default) this uses interpolation to replace `nan` diffusivity
        values. When False, `nan` values will persist into the output.

    Returns
    -------
    params : dict
        A dictionary of the extracted parameters from each pulse. The keys are
        `xs [-]` for the intercalation fractions, `Ds [m2/s]` for diffusivity,
        `i0 [A/m2]` for exchange current density, and `OCV [V]` for the OCV.
    stats : dict
        Only returned when 'return_stats' is True. Provides key/value pairs for
        the number of pulses, average pulse current, and average rest and pulse
        times.

    Raises
    ------
    ValueError
        'options' contains invalid key/value pairs.

    """

    # Options
    options = options.copy()

    R2_lim = options.pop('R2_lim', 0.95)
    replace_nans = options.pop('replace_nans', True)
    xs_ref = options.pop('xs_ref', 1.)
    ref_location = options.pop('ref_location', 'upper')

    if len(options):
        invalid_keys = list(options.keys())
        raise ValueError("'options' contains invalid key/value pairs:"
                         f" {invalid_keys=}")

    if ref_location not in ['lower', 'upper']:
        raise ValueError(f"Invalid {ref_location=}, must be 'lower', 'upper'.")

    # Pull arrays from data
    time = data.time.copy()
    current = data.current.copy()
    voltage = data.voltage.copy()

    # Constants
    R = 8.314e3  # Gas constant [J/kmol/K]
    F = 96485.33e3  # Faraday's constant [C/kmol]

    # Find pulse indexes
    if pulse_sign == 1:
        I_pulse = np.mean(current[current > 0.])
    elif pulse_sign == -1:
        I_pulse = np.mean(current[current < 0.])

    start, stop = data.find_pulses(pulse_sign)

    if start.size != stop.size:
        raise ValueError("Size mismatch: The number of detected pulse"
                         f" starts ({start.size}) and stops ({stop.size})"
                         " do not agree. This typically occurs due to a"
                         " missing rest. You will likely need to manually"
                         " remove affected pulse(s).")

    # Extract OCV
    OCV = voltage[start]

    # Extract diffusivity
    xs = np.zeros_like(OCV)
    for i in range(xs.size):
        delta_capacity = trapezoid(
            x=time[start[i]:stop[i] + 1] / 3600.,
            y=current[start[i]:stop[i] + 1] / cell.mass_AM,
        )

        if i == 0:
            xs[i] = 1.
        else:
            xs[i] = xs[i-1] - delta_capacity / cell.spec_capacity_AM

    if ref_location == 'lower':
        xs = xs - xs.min() + xs_ref
    elif ref_location == 'upper':
        xs = xs - xs.max() + xs_ref

    dOCV_dxs = np.gradient(OCV) / np.gradient(xs)

    dV_droot_t = np.zeros_like(xs)
    shifts = np.zeros(xs.size, dtype=int)
    for i in range(xs.size):
        shift = np.ceil(0.25*(stop[i] - start[i] + 1)).astype(int)

        t_pulse = time[start[i] + shift:stop[i] + 1] - time[start[i]]
        V_pulse = voltage[start[i] + shift:stop[i] + 1]

        root_t = np.sqrt(t_pulse)

        result = linregress(root_t, V_pulse)
        slope = result.slope

        while abs(result.rvalue**2) < R2_lim:
            if shift + 1 <= np.floor(0.5*(stop[i] - start[i] + 1)):
                shift += 1

                t_pulse = time[start[i] + shift:stop[i] + 1] - time[start[i]]
                V_pulse = voltage[start[i] + shift:stop[i] + 1]

                root_t = np.sqrt(t_pulse)

                result = linregress(root_t, V_pulse)
                slope = result.slope
            else:
                slope = np.nan
                break

        shifts[i] = shift
        dV_droot_t[i] = slope

    Ds = 4./np.pi * (I_pulse*cell.molar_vol_AM / (cell.surf_area_AM*F))**2 \
        * (dOCV_dxs/dV_droot_t)**2

    if any(np.isnan(Ds)) and replace_nans:
        nan = np.isnan(Ds)

        if xs[0] > xs[-1]:
            x, y = np.flip(xs[~nan]), np.flip(Ds[~nan])
        else:
            x, y = xs[~nan], Ds[~nan]

        interpolator = Akima1DInterpolator(
            x, y, method='makima', extrapolate=True,
        )

        Ds[nan] = interpolator(xs[nan])

    # Extract exchange current density
    eta_ct = voltage[start + shifts] - voltage[start]
    i0 = (R*data.avg_temperature / F) * (I_pulse / (eta_ct*cell.surf_area_AM))

    # Store output(s)
    params = pd.DataFrame({
        'xs [-]': xs,
        'Ds [m2/s]': Ds,
        'i0 [A/m2]': i0,
        'OCV [V]': OCV,
    })

    params.sort_values(by='xs [-]', inplace=True, ignore_index=True)

    stats = {
        'num pulses': start.size,
        'avg I_pulse [A]': I_pulse,
        'avg i_pulse [A/m2]': I_pulse / cell.area_ed,
        'avg t_pulse [s]': np.mean(time[stop] - time[start]),
        'avg t_rest [s]': np.mean(time[start[1:]] - time[stop[:-1]]),
    }

    if return_stats:
        return params, stats
    else:
        return params



import dash
import numpy as np
import pandas as pd
from dash import dcc, html, Output, Input, State
from scipy._lib._util import _RichResult as RichResult

from ampworks.dqdv.gui_files.pages.figures import figure_div
from ampworks.dqdv.gui_files.pages.components import (
    sliders, optimize_btns, page_spinner, terminal, logging_btns, log_toast,
    download, logging_table,
)

dash.register_page(
    __name__,
    path='/dqv-analysis',
    title='DVQ Analysis',
    page_components=[
        figure_div,
        sliders,
        optimize_btns,
        page_spinner,
        terminal,
        logging_btns,
        log_toast,
        download,
        logging_table,
    ],
)

# Page layout
layout = html.Div()

# Callbacks


@dash.callback(
    Output('terminal-out', 'children'),
    Input('summary-store', 'data'),
)
def update_terminal(summary):
    if summary:
        summary['message'] = f"'{summary.get('message', 'Done searching.')}'"
        summary['success'] = summary.get('success', True)
        summary['x'] = np.round(summary['x'], 10)
        summary['niter'] = summary.get('niter')
        summary['x_map'] = summary.pop('x_map')
        result = RichResult(**summary)
    else:
        result = RichResult({
            'message': None,
            'success': None,
            'fun': None,
            'x': None,
            'nfev': None,
            'niter': None,
            'x_map': ['x0_neg', 'x100_neg', 'x0_pos', 'x100_pos', 'iR'],
        })

    return f"```python\n{result!r}\n```"


@dash.callback(
    Output('ag-grid', 'rowData'),
    Output('log-toast', 'is_open'),
    Output('log-toast', 'children'),
    Input('add-row-btn', 'n_clicks'),
    State('ag-grid', 'rowData'),
    State('summary-store', 'data'),
    State({'type': 'filename', 'index': 'cell'}, 'children'),
    prevent_initial_call=True,
)
def log_new_row(_, current_data, summary, filename):
    if not current_data:
        current_data = []

    if not summary:
        return current_data, dash.no_update, dash.no_update

    if 'iR' not in summary['x_map']:
        summary['x'] = np.hstack([summary['x'], 0.])
        summary['x_map'].append('iR')

    new_row = dict((k, v) for k, v in zip(summary['x_map'], summary['x']))
    new_row['fun'] = summary['fun']

    new_row['filename'] = filename.removesuffix('.csv')

    current_data.append(new_row)

    return current_data, True, filename + ' added to log.'


@dash.callback(
    Output('ag-grid', 'deleteSelectedRows'),
    Input('delete-btn', 'n_clicks'),
    prevent_initial_call=True,
)
def selected(_):
    return True


@dash.callback(
    Output('download-csv', 'data'),
    Input('export-btn', 'n_clicks'),
    State('ag-grid', 'rowData'),
    State('ag-grid', 'columnDefs'),
    prevent_initial_call=True,
)
def export_to_csv(_, row_data, column_defs):
    if row_data:
        df = pd.DataFrame(row_data)
        ordered_cols = [c['field'] for c in column_defs]
        df = df[ordered_cols]
        return dcc.send_data_frame(df.to_csv, 'DVQ_Data.csv', index=False)
    else:
        return dash.no_update
